#pragma once
// Text file I/O - Unicode/ASCII/ANSI/UTF-8

#define IDS_FILE_READ_ERROR             20001
#define IDS_FILE_WRITE_ERROR            20002
#define IDS_CANNOT_OPEN_FILE            20003
#define IDS_CANNOT_OPEN_FILE_FOR_WRITING    20004

#define _CRT_SECURE_NO_WARNINGS

#define WIN32_LEAN_AND_MEAN
#include "windows.h"

#include "tchar.h"
#include "wchar.h"
#include "malloc.h"
#include "string.h"
#include "assert.h"



#define base_strlen(s)  strlen (s)
#define Hnew            new
#define tsizeof(x)      (sizeof (x) / sizeof (TCHAR))
#define T2B(x)          ((x) * sizeof (TCHAR))

// Format a string.  Caller must free
// Use MAKEINTRESOURCE to pass the ID of a string resource in message.
TCHAR * vFormatPString (const TCHAR *message, va_list ap);
TCHAR * _cdecl FormatPString (const TCHAR *message, ...);

// Return a copy of string s.  Caller must free when done
TCHAR * copy_string (const TCHAR *s);

// Make a copy of an ANSI string.  Caller must free
char * AtoA (const char *a);


// -----------------------------------------------------------------------------
// Allocation functions for growable blocks

// Allocate memory in 1k chunks.  Stores the data and allocated lengths
// immediately before the data (and returns a pointer to the data part).
void * alloc_block (void *current_block, int datalen);

// Free a block of data allocated with alloc_block
void free_block (void *block);

// Copy a block of data allocated with alloc_block
void * copy_block (const void *block);

// Return the length (not the allocated size) of a block of data allocated with alloc_block
int get_blocklen (const void *block);

// Return the amount of space allocated to this block
int get_buflen (const void *block);

// ----------------------------------------------------------------------------
// class TextFile

// filemode
#define TF_READ         1                  // existing file
#define TF_WRITE        2                  // new file
#define TF_APPEND       3                  // append to existing file

// encoding
#define TF_ANSI         1
#define TF_UTF16LE      2
#define TF_UTF16BE      3
#define TF_UTF8         4

#define TF_UNIX         0x100               // write out line-ends as \n, not \r\n
#define TF_NO_BOM_CHECK 0x200               // don't check for a BOM if explicit encoding
// specified when opening a fiole for TF_READ access

// Error codes for ReadByte / ReadChar functions (negative)
#define TF_EOF          -1
#define TF_ERROR        -2

// Conversion between UTF-16 surrogate pairs and UTF-32
// See: http://unicode.org/faq/utf_bom.html#35
// Valid lead-in codes are in the range 0xD800 to 0xDBFF inclusive
// Valid trail codes are in the range 0xDC00 to 0xDFFF inclusive
#define LEAD_OFFSET             (0xD800 - (0x10000 >> 10))
#define SURROGATE_OFFSET        (0x10000 - (0xD800 << 10) - 0xDC00)
#define C_TO_LEAD(c)            (LEAD_OFFSET + ((c) >> 10))
#define C_TO_TRAIL(c)           (0xDC00 + ((c) & 0x3FF))
#define SP_TO_C(lead, trail)    (((lead) << 10) + (trail) + SURROGATE_OFFSET)


class TextFileReader;
class TextFileWriter;
class ANSI_TextFileReader;


class TextFile
{
public:

	// Constructor
	TextFile ();

	// Destructor
	~TextFile ();

	// Open (not in constructor, lest it fail); sets last error code and string
	int Open (const TCHAR *filename, int filemode, int encoding = 0, int bufsize = 0);

	// Close; sets last error code and string if can't flush buffer
	int Close (void);

	// Flush our buffer; sets last error code and string on error
	int Flush (void);

	// ReadLine; line_buf is allocated with alloc_block and may initially be NULL
	// Character codes exceeding 0xFFFF in UTF-8 files are converted to surrogate pairs
	int ReadLine (BOOL *data_lost, TCHAR **line_buf);

	// Read a character.  Note that, for unicode files, this reads a unicode character
	// This routine correctly handles character codes exceeding 0xFFFF in UTF-8 files
	// but does not convert them to surrogate pairs.  To do this, pass the value returned
	// to C_TO_LEAD and C_TO_TRAIL if it exceeds 0xFFFF
	int ReadChar (void);

	// Write a string; sets last error code and string on error
	// Handles surrogate pairs correctly, even when writing to UTF-8 files
	// Invalid surrogate pairs are output as 0xFFFD to avoid generating a
	// non-conformant UTF-8 file
	int WriteString (BOOL *data_lost, const TCHAR *s);

	// printf, effectively; sets last error code and string on error
	int FormatString (BOOL *data_lost, const TCHAR *s, ...);

	// printf, effectively; sets last error code and string on error
	int vFormatString (BOOL *data_lost, const TCHAR *s, va_list ap);

	// Write a character; sets last error code and string on error
	// Correctly handles character codes > 0xFFFF when writing to UTF-8 files, but
	// the caller must first convert a surrogate pair to UTF-32 (by calling SP_TO_C)
	int WriteChar (BOOL *data_lost, int ch);

	// Return the file encoding (for TF_READ)
	int GetEncoding (void) { return this->encoding; }

	// Set code page; can do this without opening the file
	void SetCodePage (int code_page) { this->code_page = code_page; }

	// Set 'unknown' character; can do this without opening the file
	void SetUnknownChar (const char *unknown_char);

	// Return ANSI characters read from file unchanged
	// For ANSI-files only.
	// Note that ReadLine then returns chars, not TCHAR's
	void SetAnsiPassThru (bool passthru);

	// Return ANSI characters read from file unchanged (but widened)
	// For ANSI-files only.
	inline bool GetAnsiPassThru (void) { return this->ansi_passthru; }

private:

	// Create textfile reader and read BOM
	int CreateFileReader (void);

	// Create textfile writer
	void CreateFileWriter (void);

private:
	TCHAR  *filename;
	int     filemode;
	int     encoding;
	int     encoding_flags;
	bool    ansi_passthru;
	int     bufsize;
	HANDLE  hFile;
	TextFileReader *file_reader;
	TextFileWriter *file_writer;
	bool    need_bom;
	int     code_page;
	char   *unknown_char;
};
